#include <ctype.h>
#include <direct.h>
#include <dos.h>
#include <io.h>
#include <process.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "spawno.h"
#include "x_types.h"
#include "x_alloc.hpp"
#include "x_dv.hpp"
#include "x_cmdlin.hpp"
#include "x_crc32.hpp"
#include "x_comm.hpp"
#include "x_err.hpp"
#include "x_text.hpp"

#define PKT_AREA                0x00
#define PKT_PROMPT		0x01
#define PKT_PSTART		0x02
#define PKT_PEND		0x03
#define	PKT_MSG			0x04
#define	PKT_START		0x05
#define PKT_EXIT		0x06
#define MAX_LINES		4
#define MAX_PACKET		256
#define TIMEOUT_PACKET		2
#define COLOR_TITLE_FG		WHITE
#define COLOR_TITLE_BG		RED
#define COLOR_MSGW_FG		LIGHTGRAY
#define COLOR_MSGW_BG		BLACK
#define COLOR_CMDW_FG		0x00
#define COLOR_CMDW_BG		0x01
#define COLOR_PROMPT_FG		0x02
#define COLOR_BOLD1_FG		0x03
#define COLOR_BOLD2_FG		0x04
#define COLOR_WIN_TITLE_FG	0x05
#define COLOR_WIN_TITLE_BG	0x06
#define COLOR_WIN_FRAME_FG	0x07
#define COLOR_WIN_FRAME_BG	0x08
#define COLOR_WIN_FG		0x09
#define COLOR_WIN_BG		0x0a
#define COLOR_WIN_BOLD1_FG	0x0b
#define COLOR_WIN_BOLD2_FG	0x0c
#define N_COLORS		0x0d

void		init();
void		SIGINT_handler(int sig);
void		deInit();
void		screenInit();
void		commandScreenInit();
void		msgW();
void		cmdW();
void		processConfig();
int		getToken(char *d, char *s, int *idx);
int		getString(char *d, char *s, int *idx);
void		modemInit();
void		connect();
void		login();
void            runScript(char *path);
void		title(char *s, ...);
int		expect(char *s, int timeout);
int		expect2(char *s1, char *s2, int timeout);
int		command();
int		getPacket(int *len, byte *data);
void		beep();
void		msgStr(char *s);
void            waitKey();
int		terminal();
void		shell();
void		helpWindow();
void		createWindow(int l, int t, int r, int b, char *title);
void		destroyWindow();
void		ZModemUpload();
void		ZModemDownload();

static const int		stdBase[] = { 0x3f8, 0x2f8, 0x3e8, 0x2e8 };
static const int		stdIRQ[] = { 4, 3, 4, 3 };
static const long		validBPS[] = { 14400, 19200, 28800, 38400, 57600, 115200, 0 };
static const char		*configColorName[N_COLORS] = {
	"CMDW_FG", "CMDW_BG", "PROMPT_FG", "BOLD1_FG", "BOLD2_FG",
	"WIN_TITLE_FG", "WIN_TITLE_BG", "WIN_FRAME_FG", "WIN_FRAME_BG",
	"WIN_FG", "WIN_BG", "WIN_BOLD1_FG", "WIN_BOLD2_FG"
};
static const char		*colorName[] = { "Black", "Blue", "Green",
	"Cyan", "Red", "Magenta", "Brown", "LightGray", "DarkGray",
	"LightBlue", "LightGreen", "LightCyan", "LightRed",
	"LightMagenta", "Yellow", "White"
};
static const char		titleString[] = "LGC-Comm v1.02 by Adam McKee";

static int			color[N_COLORS] = {
	WHITE, BLUE, YELLOW, GREEN, LIGHTGRAY, WHITE, BLACK, BLUE, BLACK,
	LIGHTGRAY, BLACK, YELLOW, GREEN
};
static char			username[16];
static char			password[16];
static char			ulPath[128];
static char			dlPath[128];
static char			modemInitStr[64];
static char			modemDialStr[64];
static int			nPhNos;
static char			phNo[32][11];
static char			nick[32];
static char			*lgcStartStr = "___LGC_START___";
static int			lgcStartStrN, lgcStartStrLen = strlen(lgcStartStr);
static char			ZModemStart[] = { 0x42, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x00 };
static int			ZModemStartN, ZModemStartLen = strlen(ZModemStart);
static int			port;
static int			base;
static int			irq;
static dword			bps;
static int			curW, cmdWx = 1, cmdWy = 1,
				msgWx = 1, msgWy = 1;
static int			rows;
static int                      paging, pagingLines;
static int			hangup = 1;
static int			terminalRunning;
static word			**savedScreen;
static text_info		savedTextinfo;
static short                    curArea;

void main(int argc) {
  init();
  atexit(deInit);
  if (argc == 1) {
    textcolor(COLOR_TITLE_FG);
    textbackground(COLOR_TITLE_BG);
    clreol();
    gotoxy((80 - strlen(titleString)) / 2, 1);
    cprintf("%s\r\n", titleString);
    window(1, 2, 80, rows);
    textbackground(BLACK);
    title("Alt-H for Online Help");
    title("Using COM%d (%s,0x%x,%d) @ %lu BPS", port + 1, comm->UART_16550() ? "16550A" : "8250/16450", base, irq, bps);
    comm->lowerDTR(500);
    if (modemInitStr[0] && modemDialStr[0]) {
      modemInit();
      connect();
    } else comm->write("\r\r");
    login();
  } else {
    for (int i = 0; i < 16; i++) {
      comm->write(0);
      comm->write(0xff);
    }
    if (command()) exit(255);
  }
  for (;;) {
    if (terminal()) break;
    screenInit();
    if (nick[0]) {
      delay(1000);
      comm->write("/nick "); comm->write(nick); comm->write('\r');
    }
    if (command()) break;
  }
  exit(255);
}

void init() {
/*  FILE *f = fopen("telix.cap", "rb");
  int c;
  while ((c = fgetc(f)) != EOF) {
    delay(10);
    wputch(c);
  }
  fclose(f);
  exit(0); */
  processConfig();
  screenInit();
  comm = new Comm(base, irq, bps, 3, 0);
  signal(SIGINT, SIGINT_handler);
}

void deInit() {
  if (comm) {
    if (hangup) {
      if (!modemInitStr[0] && !modemDialStr[0]) comm->write(4);
      comm->lowerDTR(500);
    }
    delete comm;
  }
  _setvideomode(_TEXTC80); rows = 25; screenInit();
}

void SIGINT_handler(int sig) {
  sig = sig;
  if (terminalRunning) {
    comm->write(3);
    signal(SIGINT, SIGINT_handler);
  } else exit(255);
}

void screenInit() {
  if (!rows) {
    _setvideomode(_TEXTC80); rows = 25;
  }
  window(1, 1, 80, rows);
  textcolor(LIGHTGRAY);
  textbackground(BLACK);
  clrscr();
}

void commandScreenInit() {
  msgW();
  clrscr();
  cmdW();
  clrscr();
}

void msgW() {
  if (curW == 0) return;
  cmdWx = wherex(); cmdWy = wherey();
  window(1, 1, 80, rows - 1);
  textcolor(COLOR_MSGW_FG);
  textbackground(COLOR_MSGW_BG);
  gotoxy(msgWx, msgWy);
  curW = 0;
}

void cmdW() {
  if (curW == 1) return;
  msgWx = wherex(); msgWy = wherey();
  window(1, rows, 80, rows);
  textcolor(color[COLOR_CMDW_FG]);
  textbackground(color[COLOR_CMDW_BG]);
  gotoxy(cmdWx, cmdWy);
  curW = 1;
}

void processConfig() {
  FILE *f = fopen("lgc.cfg", "rt");
  int i, j, li;
  char line[256], token[64];
  if (f == NULL) fatalError("No configuration file!\n", 255);
  while (fgets(line, 256, f) != NULL) {
    li = 0;
    if (!getToken(token, line, &li)) {
      if (token[0] == '#') continue;
      else if (!stricmp(token, "TEXTMODE")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
        if (!stricmp(token, "80x25")) {
          _setvideomode(_TEXTC80);
          rows = 25;
        } else if (!stricmp(token, "80x50")) {
          _settextrows(50);
          rows = 50;
        }
      } else if (!stricmp(token, "USERNAME")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(username, line, &li)) continue;
      } else if (!stricmp(token, "PASSWORD")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(password, line, &li)) continue;
      } else if (!stricmp(token, "UL_PATH")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(ulPath, line, &li)) continue;
      } else if (!stricmp(token, "DL_PATH")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(dlPath, line, &li)) continue;
      } else if (!stricmp(token, "MODEM_INIT")) {
	if (getToken(token, line, &li)) continue;
	if (getString(modemInitStr, line, &li)) continue;
      } else if (!stricmp(token, "MODEM_DIAL")) {
	if (getToken(token, line, &li)) continue;
	if (getString(modemDialStr, line, &li)) continue;
      } else if (!stricmp(token, "NICK")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	strncpy(nick, token, 16);
	nick[16] = '\0';
      } else if (!stricmp(token, "PH_NO")) {
	if (getToken(token, line, &li)) continue;
	while (!getToken(token, line, &li) && nPhNos < 32) {
	  if (strlen(token) > 10) continue;
	  strcpy(phNo[nPhNos++], token);
	}
      } else if (!stricmp(token, "PORT")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	port = atoi(token) - 1;
	if (port < 0 || port > 3) fatalError("Illegal port: %d", 255, port + 1);
	base = stdBase[port];
	irq = stdIRQ[port];
      } else if (!stricmp(token, "BASE")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	sscanf(token, "%x", &base);
      } else if (!stricmp(token, "IRQ")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	irq = atoi(token);
      } else if (!stricmp(token, "BPS")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	bps = atol(token);
	for (i = 0; validBPS[i]; i++) if (bps == validBPS[i]) break;
	if (validBPS[i] == 0) fatalError("Unknown BPS: %s", 255, token);
      } else {
	for (i = 0; i < N_COLORS; i++) {
	  if (!stricmp(token, configColorName[i])) {
	    if (getToken(token, line, &li)) continue;
	    if (getToken(token, line, &li)) continue;
	    for (j = 0; j < 16; j++) {
	      if (!stricmp(token, colorName[j])) {
		color[i] = j;
	      }
	    }
	  }
	}
      }
    }
  }
  fclose(f);
}

int getToken(char *d, char *s, int *idx) {
  int si, di, sLen = strlen(s);
  for (si = *idx; si < sLen && isspace(s[si]); si++);
  for (di = 0; si < sLen && isgraph(s[si]); si++, di++) d[di] = s[si];
  d[di] = '\0';
  *idx = si;
  return di == 0;
}

int getString(char *d, char *s, int *idx) {
  int si, di, sLen = strlen(s);
  for (si = *idx; si < sLen && s[si] != '"'; si++);
  si++;
  for (di = 0; si < sLen && s[si] != '"'; si++, di++) d[di] = s[si];
  si++;
  d[di] = '\0';
  *idx = si;
  return di == 0;
}

void modemInit() {
  char cmd[64], res[64];
  textcolor(color[COLOR_BOLD1_FG]);
  title("Initializing Modem");
  comm->command("AT\r", res);
  textcolor(color[COLOR_BOLD2_FG]);
  sprintf(cmd, "%s\r", modemInitStr);
  comm->command(cmd, res);
  cprintf("%s", res);
}

void connect() {
  int c, i, curPhNo = 0, hangup = 0;
  char cmd[64], res[64];
  while (kbhit()) getch();
l0:
  if (hangup) {
    hangup = 0;
    comm->lowerDTR(500);
  }
  delay(2000);
  textcolor(color[COLOR_BOLD1_FG]);
  title("Dialing");
  textcolor(color[COLOR_BOLD2_FG]);
  sprintf(cmd, "%s%s\r", modemDialStr, phNo[curPhNo]);
  if (++curPhNo == nPhNos) curPhNo = 0;
  comm->command(cmd, res);
  cprintf("%s\n", res);
  i = expect2("CONNECT", "BUSY", 45);
  if (!i) {
    if (!kbhit()) {
      hangup = 1;
      goto l0;
    }
    while (kbhit()) c = getch();
    if (c == ' ') {
      hangup = 1;
      goto l0;
    } else exit(255);
  } else if (i == 2) goto l0;
}

void login() {
  textcolor(color[COLOR_BOLD1_FG]);
  title("Logging In");
  delay(2000);
  if (!access("login.scr", 0)) runScript("login.scr");
  else {
    comm->write('\r');
    comm->write('\r');
    if (!expect2("ogin:", "sername:", 10)) exit(255);
    comm->write(username); comm->write('\r');
    if (!expect("assword:", 10)) exit(255);
    comm->write(password); comm->write('\r');
  }
}

void runScript(char *path) {
  FILE *f;
  int li, res;
  char line[256], token[256], token2[256], *s;
  f = fopen(path, "rt");
  if (!f) exit(255);
  while (fgets(line, 256, f) != NULL) {
    li = 0;
    if (!getToken(token, line, &li)) {
      if (token[0] == '#') continue;
      else if (!stricmp(token, "EXPECT")) {
	if (getToken(token, line, &li)) continue;
        if (!getToken(token2, line, &li)) res = expect(token, atoi(token2));
        else res = expect(token, 10);
        if (!res) fatalError("EXPECT failed", 255);
      } else if (!stricmp(token, "SEND")) {
        if (getToken(token, line, &li)) continue;
        s = token;
        while (*s) {
          if (*s == '\\') {
            switch (*++s) {
              case '\\': comm->write('\\'); s++; break;
              case 'r': comm->write('\r'); s++; break;
              case 'n': comm->write('\n'); s++; break;
              default: comm->write('\\'); comm->write(*s++); break;
            }
          } else if (*s == '$') {
            s++;
            if (!strnicmp(s, "USERNAME", 8)) {
              comm->write(username);
              s += 8;
            } else if (!strnicmp(s, "PASSWORD", 8)) {
              comm->write(password);
              s += 8;
            } else comm->write('$');
          } else comm->write(*s++);
        }
      } else if (!stricmp(token, "SLEEP")) {
	if (getToken(token, line, &li)) continue;
        delay(atoi(token));
      }
    }
  }
  fclose(f);
}

void title(char *fmt, ...) {
  char buf[80];
  textcolor(WHITE);
  va_list argptr;
  va_start(argptr, fmt);
  vsprintf(buf, fmt, argptr);
  cprintf("\r\n");
  cprintf("[ ");
  textcolor(color[COLOR_BOLD1_FG]);
  cprintf(buf);
  textcolor(WHITE);
  cprintf(" ]");
  cprintf("\r\n");
  va_end(argptr);
}

int expect(char *s, int timeout) {
  int matchLen = 0, sLen = strlen(s);
  byte c;
  time_t t1, t2;
  time(&t1);
  for (;;) {
    while (comm->iEmpty()) {
      time(&t2);
      if (t2 - t1 >= timeout || kbhit()) return 0;
    }
    comm->read((char*)&c);
    if (c == s[matchLen]) {
      matchLen++;
      if (matchLen == sLen) return 1;
    } else matchLen = 0;
  }
}

int expect2(char *s1, char *s2, int timeout) {
  int matchLen1 = 0, matchLen2 = 0,
      s1_len = strlen(s1), s2_len = strlen(s2);
  byte c;
  time_t t1, t2;
  time(&t1);
  for (;;) {
    while (comm->iEmpty()) {
      time(&t2);
      if (t2 - t1 >= timeout || kbhit()) return 0;
    }
    comm->read((char*)&c);
    if (c == s1[matchLen1]) {
      matchLen1++;
      if (matchLen1 == s1_len) return 1;
    } else matchLen1 = 0;
    if (c == s2[matchLen2]) {
      matchLen2++;
      if (matchLen2 == s2_len) return 2;
    } else matchLen2 = 0;
  }
}

int command() {
  int res = 0, i, c, pktLen, runCmd = 0, len = 0, nLines = 0;
  curArea = -1;
  FILE *f;
  char *prompt, *line[MAX_LINES], *cmdline;
  byte *pktData;
  prompt = (char*)smalloc(32); prompt[0] = '\0';
  for (i = 0; i < MAX_LINES; i++) line[i] = (char*)smalloc(81);
  pktData = (byte*)smalloc(MAX_PACKET);
  cmdline = (char*)smalloc(MAX_LINES * 80 + 1);
  commandScreenInit();
  for (;;) {
    if (kbhit()) {
      c = getch();
      if (!c) {
	c = getch();
	switch (c) {
	  case 20:
	    goto l0;
	  case 31:
            if (curArea >= 0) {
	      comm->write("/me shelled to DOS\r");
              while (!comm->oEmpty()); delay(250);
	      while (comm->iEmpty());
              while (!comm->iEmpty())
                if (!getPacket(&pktLen, pktData))
                  if (pktData[0] == PKT_MSG)
	            msgStr((char*)pktData + 1);
            }
	    shell();
            delay(100); while (!comm->iEmpty()) getPacket(&pktLen, pktData);
            if (curArea >= 0)
	      comm->write("/me has returned\r");
	    comm->write("/list\r");
	    break;
	  case 35:
	    helpWindow();
	    break;
	  case 45:
	    res = 1;
	    goto l0;
	  default:
	    continue;
	}
      } else if (isprint(c)) {
	line[nLines][len++] = c;
	cprintf("%c", c);
      } else if (c == 1) {
	textcolor(color[COLOR_PROMPT_FG]);
	cprintf(prompt);
	textcolor(color[COLOR_CMDW_FG]);
	nLines = len = 0;
      } else if (c == 8) {
	if (len) {
	  cmdWx = wherex() - 1; cmdWy = wherey();
	  gotoxy(cmdWx, cmdWy);
	  cprintf(" ");
	  gotoxy(cmdWx, cmdWy);
	  len--;
	} else if (nLines) {
	  line[nLines--][0] = '\0';
	  len = strlen(line[nLines]);
	  clrscr();
	  cprintf(line[nLines]);
	} else beep();
      } else if (c == 13) {
	runCmd = 1;
      } else if (c == 24) {
	comm->write(24);
      } else if (isspace(c)) {
	line[nLines][len++] = ' ';
	cprintf(" ");
      } else beep();
      if (len && !(len % 80)) {
	if (nLines < MAX_LINES - 1) {
	  line[nLines][len] = '\0';
	  for (i = len; i >= 0 && line[nLines][i] != ' '; i--);
	  if (i >= 0) {
	    strcpy(line[nLines + 1], line[nLines] + i + 1);
	    while (line[nLines][i - 1] == ' ' && i) i--;
	    line[nLines][i] = '\0';
	  } else {
	    line[nLines + 1][0] = line[nLines][len - 1];
	    line[nLines + 1][1] = '\0';
	    line[nLines][len - 1] = '\0';
	  }
	  nLines++;
	  len = strlen(line[nLines]);
	  clrscr();
	  cprintf(line[nLines]);
	} else runCmd = 1;
      } else if (prompt[0] && len == 16) runCmd = 1;
      if (runCmd) {
	runCmd = 0;
	line[nLines][len] = '\0';
	strcpy(cmdline, line[0]);
	for (i = 1; i <= nLines; i++) {
	  if (line[i][0] != ' ') strcat(cmdline, " ");
	  strcat(cmdline, line[i]);
	}
	for (i = 0; i < 3; i++) comm->write(1);
	comm->write(cmdline); comm->write('\r');
	clrscr();
	textcolor(color[COLOR_PROMPT_FG]);
	cprintf(prompt);
	textcolor(color[COLOR_CMDW_FG]);
	nLines = len = 0;
      }
    }
    if (!comm->iEmpty()) {
      if (getPacket(&pktLen, pktData)) continue;
      switch (pktData[0]) {
        case PKT_AREA:
          curArea = *(short*)(pktData + 1);
          break;
	case PKT_PROMPT:
	  strcpy(prompt, (char*)pktData + 1);
	  if (!len) {
	    clrscr();
	    textcolor(color[COLOR_PROMPT_FG]);
	    cprintf(prompt);
	    textcolor(color[COLOR_CMDW_FG]);
	  }
	  break;
        case PKT_PSTART:
          paging = 1;
          pagingLines = 0;
          break;
        case PKT_PEND:
          paging = 0;
          break;
	case PKT_MSG:
	  msgStr((char*)pktData + 1);
	  break;
	case PKT_START:
	  hangup = 0;
	  f = fopen("lgc.tmp", "wb");
	  fputs((char*)pktData + 2, f);
	  fclose(f);
	  exit(pktData[1]);
	  break;
	case PKT_EXIT:
	  res = 0;
	  goto l1;
      }
    }
  }
l0:
  comm->write("/quit\r");
  while (comm->iEmpty());
  getPacket(&pktLen, pktData);
l1:
  sfree(prompt);
  for (i = 0; i < MAX_LINES; i++) sfree(line[i]);
  sfree(pktData);
  sfree(cmdline);
  return res;
}

int getPacket(int *len, byte *data) {
  int esc = 0;
  byte c;
  time_t t1, t2;
  *len = 0;
  for (;;) {
    comm->read((char*)&c);
    if (esc) {
      esc = 0;
      if (c == 0xff) {
        data[(*len)++] = c;
        if (*len == MAX_PACKET) return 1;
      } else {
        if (*len < sizeof(dword)) return 1;
        *len -= sizeof(dword);
        CRC32 crc;
        for (int i = 0; i < *len; i++)
          crc.addByte(data[i]);
        if (crc.CRC() != *(dword*)(data + *len)) {
          if (curArea >= 0) {
            msgStr("\n");
            comm->write("/me dropped a packet\r");
          }
          return 1;
        }
	data[(*len)++] = '\0';
	return 0;
      }
    } else {
      if (c == 0xff) esc = 1;
      else {
        data[(*len)++] = c;
        if (*len == MAX_PACKET) return 1;
      }
    }
    time(&t1);
    while (comm->iEmpty()) {
      time(&t2);
      if (t2 - t1 >= TIMEOUT_PACKET) return 1;
    }
  }
}

void beep() {
  sound(500);
  delay(5);
  nosound();
}

void msgStr(char *s) {
  int i, l = strlen(s);
  msgW();
  for (i = 0; i < l; i++) {
    if (s[i] == '\n') {
      if (paging && ++pagingLines == rows - 1) {
        waitKey();
        pagingLines = 0;
      }
    }
    wputch(s[i]);
  }
  cmdW();
}

void waitKey() {
  cmdW();
  clrscr();
  gotoxy(23, 1);
  cprintf(">>> Press any key to continue <<<");
  while (!kbhit()); while (kbhit()) getch();
  clrscr();
  msgW();
}

int terminal() {
  int res = 0, c;
  char ansiStr[64], c2;
  terminalRunning = 1;
  screenInit();
  for (;;) {
l0:
    if (kbhit()) {
      c = getch();
      if (!c) {
	c = getch();
	switch (c) {
	  case 31:
	    shell();
	    break;
	  case 35:
	    helpWindow();
	    break;
	  case 45:
	    res = 1;
	    goto l1;
	  case 72:
	    comm->write("\x1b[A");
	    break;
	  case 73:
	    ZModemUpload();
	    break;
	  case 75:
	    comm->write("\x1b[D");
	    break;
	  case 77:
	    comm->write("\x1b[C");
	    break;
	  case 80:
	    comm->write("\x1b[B");
	    break;
	}
      } else {
	if (c == 8) comm->write(127);
	else comm->write(c);
      }
    }
    if (comm->getIcount()) {
      comm->read(&c2);
      wputch(c2);
      if (ansiDimensionsRequested) {
        sprintf(ansiStr, "\x1b[%d;80R", rows);
	comm->write(ansiStr);
	ansiDimensionsRequested = 0;
      }
      if (c2 == lgcStartStr[lgcStartStrN]) {
	if (++lgcStartStrN == lgcStartStrLen) goto l1;
      } else lgcStartStrN = 0;
      if (c2 == ZModemStart[ZModemStartN]) {
	if (++ZModemStartN == ZModemStartLen) {
	  ZModemStartN = 0;
	  ZModemDownload();
	}
      } else ZModemStartN = 0;
    }
  }
l1:
  terminalRunning = 0;
  return res;
}

void shell() {
  char lgcPath[64];
  word lgcDrive, dummy;
  delete comm;
  dv_saveScreen(&savedScreen); gettextinfo(&savedTextinfo);
  screenInit();
  getcwd(lgcPath, 64);
  _dos_getdrive(&lgcDrive);
  spawnlo(NULL, getenv("COMSPEC"), getenv("COMSPEC"), NULL);
  _dos_setdrive(lgcDrive, &dummy);
  chdir(lgcPath);
  if (rows == 25) _setvideomode(_TEXTC80);
  else _settextrows(50);
  destroyWindow();
  comm = new Comm(base, irq, bps, 3, 0);
}

void helpWindow() {
  _setcursortype(_NOCURSOR);
  createWindow(12, 3, 68, 23, "LGC-Comm Online Help");
  textcolor(color[COLOR_WIN_BOLD1_FG]); gotoxy(4, 3); cprintf("When dialing:");
  textcolor(color[COLOR_WIN_BOLD2_FG]); gotoxy(6, 5); cprintf("SPACE");
  textcolor(color[COLOR_WIN_FG]); gotoxy(12, 5); cprintf("to abort current dialing attempt");
  textcolor(color[COLOR_WIN_BOLD2_FG]); gotoxy(8, 6); cprintf("ESC");
  textcolor(color[COLOR_WIN_FG]); gotoxy(12, 6); cprintf("to exit to DOS");

  textcolor(color[COLOR_WIN_BOLD1_FG]); gotoxy(4, 8); cprintf("When in LGC mode:");
  textcolor(color[COLOR_WIN_BOLD2_FG]); gotoxy(6, 10); cprintf("Alt-T");
  textcolor(color[COLOR_WIN_FG]); gotoxy(12, 10); cprintf("to enter ANSI terminal mode");

  textcolor(color[COLOR_WIN_BOLD1_FG]); gotoxy(4, 12); cprintf("When in ANSI terminal mode:");
  textcolor(color[COLOR_WIN_BOLD2_FG]); gotoxy(7, 14); cprintf("PgUp");
  textcolor(color[COLOR_WIN_FG]); gotoxy(12, 14); cprintf("to start a ZModem upload");

  textcolor(color[COLOR_WIN_BOLD1_FG]); gotoxy(4, 16); cprintf("In either mode:");
  textcolor(color[COLOR_WIN_BOLD2_FG]); gotoxy(6, 18); cprintf("Alt-S");
  textcolor(color[COLOR_WIN_FG]); gotoxy(12, 18); cprintf("for a DOS shell");
  textcolor(color[COLOR_WIN_BOLD2_FG]); gotoxy(6, 19); cprintf("Alt-X");
  textcolor(color[COLOR_WIN_FG]); gotoxy(12, 19); cprintf("to exit to DOS");

  while (!kbhit()); while (kbhit()) getch();
  destroyWindow();
  _setcursortype(_NORMALCURSOR);
}

void createWindow(int l, int t, int r, int b, char *title) {
  int x, y;
  dv_saveScreen(&savedScreen); gettextinfo(&savedTextinfo);
  window(l, t, r, b);
  textbackground(color[COLOR_WIN_BG]);
  clrscr();
  textcolor(color[COLOR_WIN_FRAME_FG]);
  textbackground(color[COLOR_WIN_FRAME_BG]);
  for (x = 2; x < r - l + 1; x++) {
    gotoxy(x, 1);
    wputch('');
    gotoxy(x, b - t + 1);
    wputch('');
  }
  for (y = 2; y < b - t + 1; y++) {
    gotoxy(1, y);
    wputch('');
    gotoxy(r - l + 1, y);
    wputch('');
  }
  dv_WAputch(1, 1, (byte)'');
  dv_WAputch(r - l + 1, 1, (byte)'');
  dv_WAputch(1, b - t + 1, (byte)'');
  dv_WAputch(r - l + 1, b - t + 1, (byte)'');
  dv_WAputch(((r - l + 1) - (strlen(title) + 2)) / 2 - 1, 1, (byte)'');
  dv_WAputch(((r - l + 1) - (strlen(title) + 2)) / 2 + strlen(title) + 2, 1, (byte)'');
  textcolor(color[COLOR_WIN_TITLE_FG]);
  textbackground(color[COLOR_WIN_TITLE_BG]);
  gotoxy(((r - l + 1) - (strlen(title) + 2)) / 2, 1);
  cprintf(" %s ", title);
  textcolor(color[COLOR_WIN_FG]);
  textbackground(color[COLOR_WIN_BG]);
}

void destroyWindow() {
  dv_restoreScreen(savedScreen);
  window(savedTextinfo.winleft, savedTextinfo.wintop,
	 savedTextinfo.winright, savedTextinfo.winbottom);
  textattr(savedTextinfo.attribute);
  gotoxy(savedTextinfo.curx, savedTextinfo.cury);
}

void ZModemUpload() {
  char ulDrive[3], lgcPath[64], path[45], str[256];
  word lgcDrive, dummy;
  int i, c, pathLen = 0;
  createWindow(12, 7, 68, 13, "ZModem Upload");
  textcolor(color[COLOR_WIN_BOLD1_FG]); gotoxy(4, 3); cprintf("Path:");
  textcolor(color[COLOR_WIN_BOLD2_FG]); gotoxy(19, 5); cprintf("ESC");
  textcolor(color[COLOR_WIN_FG]); gotoxy(23, 5); cprintf("to abort upload");
  gotoxy(10, 3);
  for (;;) {
    while (!kbhit());
    c = getch();
    if (!c) {
      c = getch();
      continue;
    }
    if (c == 8) {
      if (pathLen) {
	gotoxy(wherex() - 1, wherey());
	wputch(' ');
	gotoxy(wherex() - 1, wherey());
	pathLen--;
      }
    } else if (c == 13) break;
    else if (c == 27) goto l0;
    else if (isprint(c)) {
      if (pathLen < 44) {
	path[pathLen++] = c;
	wputch(c);
      }
    }
  }
  path[pathLen] = '\0';
  for (i = 0; isgraph(path[i]); i++);
  if (!i) goto l0;
  path[i] = '\0';
  strncpy(ulDrive, ulPath, 2);
  ulDrive[2] = '\0';
  _dos_getdrive(&lgcDrive);
  getcwd(lgcPath, 64);
  delete comm;
  screenInit();
  sprintf(str, "ul.bat %s %s %s %x %d %lu %s", ulDrive, ulPath, lgcPath, base, irq, bps, path);
  systemo(NULL, str);
  comm = new Comm(base, irq, bps, 3, 0);
  _dos_setdrive(lgcDrive, &dummy);
  chdir(lgcPath);
  if (rows == 25) _setvideomode(_TEXTC80);
  else _settextrows(50);
l0:
  destroyWindow();
}

void ZModemDownload() {
  char dlDrive[3], lgcPath[64], str[256];
  word lgcDrive, dummy;
  strncpy(dlDrive, dlPath, 2);
  dlDrive[2] = '\0';
  _dos_getdrive(&lgcDrive);
  getcwd(lgcPath, 64);
  delete comm;
  dv_saveScreen(&savedScreen); gettextinfo(&savedTextinfo);
  screenInit();
  sprintf(str, "dl.bat %s %s %s %x %d %lu", dlDrive, dlPath, lgcPath, base, irq, bps);
  systemo(NULL, str);
  comm = new Comm(base, irq, bps, 3, 0);
  _dos_setdrive(lgcDrive, &dummy);
  chdir(lgcPath);
  if (rows == 25) _setvideomode(_TEXTC80);
  else _settextrows(50);
  destroyWindow();
  comm->write('\n');
}
